package ElchemeaDevice;
#$Id: ElchemeaDevice.pm,v 1.14 2012/03/22 14:22:52 sofc Exp $

################################################################################
#
#   Package Elchemeadevice
#
#   This class is only a wrapper for the individual frequency device classes
#   and as such does not provide any special functionality apart from the three
#   functions to return a frequency device instance.
#
#   $id = new_device($devicename) returns a new device based on the specified name 
#
#   $id = new_FRA($devicename)  returns a new frequency response analyser device
#   based on specified name. A special name is 'NONE' in which case a dummy device
#   is returned. This dummydevice honors the excact same methods as a real device
#   except that it returns dummydata.
#
#   $id = new_PSTAT($devicename) returns a new potentiostat device based on the 
#   specified name. As with new_FRA() the special 'NONE' name is recognised and
#   in this case a dummy potentiostat device is returned.
#
#   Common for both types of dummydevices is that the member function virtual() 
#   returns true (1) (it returns false (0) for all real devices).
#
#   All member functions on frequency devices are specified in the FDEV.pm class 
#   file (which all frequency device classes derive from). The FDEV class contains
#   a number of virtual functions which gets overridden by the individual device 
#   classes as well as a number of common member functions.
#
#   Two utility functions exist: get_FRA_list() and get_PSTAT_list()
#   Both functions merely returns a list of valid device names for FRA's and.
#   potentiostats respectively.
#
################################################################################

use Solartron1260;
use Solartron1255;
use Solartron1250;
use Solartron1287;
use Solartron1280;
use Hioki;
use Stanford;
use Elchemea qw ();

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[qw/new_device new_FRA new_PSTAT/]
		    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}


sub new_device($@)
{
    my $type = shift;
    my $id = undef;
    $type = '' unless defined($type);
    $_ = $type;
  CASE:
    {
        /^Solartron\s*1260$/ and do {$id = Solartron1260->new(@_); last CASE};
        /^Solartron\s*1250$/ and do {$id = Solartron1250->new(@_); last CASE};
        /^Solartron\s*1250SL$/ and do {$id = Solartron1250SL->new(@_); last CASE};
        /^Solartron\s*1255$/ and do {$id = Solartron1255->new(@_); last CASE};
        /^Solartron\s*1255A$/ and do {$id = Solartron1255->new(@_); last CASE};
        /^Solartron\s*1255B$/ and do {$id = Solartron1255B->new(@_); last CASE};
        /^Solartron\s*1252$/ and do {$id = Solartron1252->new(@_); last CASE};
        /^Solartron\s*1287$/ and do {$id = Solartron1287->new(@_); last CASE};
        /^Solartron\s*1286$/ and do {$id = Solartron1286->new(@_); last CASE};
	/^Solartron\s*1280$/ and do {$id = Solartron1280->new(@_); last CASE};
	/^Solartron\s*1280A$/ and do {$id = Solartron1280A->new(@_); last CASE};
	/^Solartron\s*1280FRA$/ and do {$id = Solartron1280_FRA->new(@_); last CASE};
	/^Hioki\s*3522$/ and do {$id = Hioki3522->new(@_); last CASE};
	/^Hioki\s*3532$/ and do {$id = Hioki3532->new(@_); last CASE};
	/^Stanford\s*SR830/ and do {$id = Stanford->new(@_); last CASE;};
    }
    die "Unknown device '$type'" unless ($id);
    my $port = Elchemea::get_config_value('equipment','port');
    $id->port($port) if ($port);
    my $host = Elchemea::get_config_value('equipment','host');
    $id->host($host) if ($host);
    return $id;
}

sub get_FRA_list()
{
    my @list = ('NONE','Solartron 1250','Solartrin 1250SL','Solartron 1260','Solartron 1255','Solartron 1255A','Solartron 1255B','Solartron 1252','Solartron 1280','Hioki 3522','Hioki 3532','Stanford SR830');
    return @list;
}

sub new_FRA($@)
{
    my $type = shift;
    my $addr = shift;
    $type = '' unless defined($type);
    my $id = undef;
    $_ = $type;
  CASE:
    {
	/^NONE$/ and do {$id = FRA->new(0,@_); last CASE};
        /^Solartron\s*1250$/ and do {$id = Solartron1250->new($addr,@_); last CASE};
        /^Solartron\s*1250SL$/ and do {$id = Solartron1250SL->new($addr,@_); last CASE};
	/^Solartron\s*1260$/ and do {$id = Solartron1260->new($addr,@_); last CASE};
        /^Solartron\s*1255$/ and do {$id = Solartron1255->new($addr,@_); last CASE};
        /^Solartron\s*1255A$/ and do {$id = Solartron1255->new($addr,@_); last CASE};
        /^Solartron\s*1255B$/ and do {$id = Solartron1255B->new($addr,@_); last CASE};
        /^Solartron\s*1252$/ and do {$id = Solartron1252->new($addr,@_); last CASE};
	/^Solartron\s*1280$/ and do {$id = Solartron1280_FRA->new($addr,@_); last CASE};
	/^Hioki\s*3522$/ and do {$id = Hioki3522->new($addr,@_); last CASE};
	/^Hioki\s*3532$/ and do {$id = Hioki3532->new($addr,@_); last CASE};
	/^Stanford\s*SR830/ and do {$id = Stanford->new($addr,@_); last CASE;};
    }
    die "Unknown device '$type'" unless ($id);
    my $port = Elchemea::get_config_value('equipment','port');
    $id->port($port) if ($port);
    my $host = Elchemea::get_config_value('equipment','host');
    $id->host($host) if ($host);
    return $id;
}

sub get_PSTAT_list()
{
    my @list = ('NONE','Solartron 1287','Solartron 1286','Solartron 1280','Solartron 1280A');
    return @list;
}

sub new_PSTAT($@)
{
    my $type = shift;
    my $addr = shift;
    my $id = undef;
    $type = '' unless defined($type);
    $_ = $type;
  CASE:
    {
	/^NONE$/ and do {$id = PSTAT->new(0,@_); last CASE};
        /^Solartron\s*1287$/ and do {$id = Solartron1287->new($addr,@_); last CASE};
        /^Solartron\s*1286$/ and do {$id = Solartron1286->new($addr,@_); last CASE};
	/^Solartron\s*1280$/ and do {$id = Solartron1280->new($addr,@_); last CASE};
	/^Solartron\s*1280A$/ and do {$id = Solartron1280A->new($addr,@_); last CASE};
    }
    die "Unknown device '$type'" unless ($id);
    my $port = Elchemea::get_config_value('equipment','port');
    $id->port($port) if ($port);
    my $host = Elchemea::get_config_value('equipment','host');
    $id->host($host) if ($host);
    return $id;
}


return 1;
