package ElchemeaUser;
#$Id: ElchemeaUser.pm,v 1.61 2016/03/03 12:46:20 sofc Exp $

################################################################################
#
# Package ElchemeaUser
# 
# This module defines the ElchemeaUser class. To obtain an ElchemeaUser instance, 
# call the constructor:
#
# my $u = ElchemeaUser->new($username);
#
# All ElchemeaUser instances has the following public member functions:
#
# $id->modtime([$])                  Returns the last modification time for the configuration file
#                                    for the specified session number (default is current session).
# $id->debug([$])                    Sets or gets the debug level.
# $id->init()                        Initialises the ElchemaUser instance (is automatically called
#                                    by the constructor, but can be called explicitly).
# $id->get_channels()                Returns a string with the data from the datalogging devices/channels
#                                    specified for the user. The string consists of a number of 'name: value'
#                                    pairs separated by ', '.
# $id->log_string()                  Returns the last user log entry from the current session.
# $id->sample_name()                 Returns the sample name for the current session.
# $id->get_measure_file()            Returns the name of the current measure file. Argument: the file name extension.
# $id->fisher_yates_shuffle()        Shuffles the supplied array. Argument a refernece to the array to be shuffled..
# $id->get_potsweep_points($)        Returns a list of potential sweep apex points based on 
#                                    specified session configuration (default is current session).
# $id->get_offset_program_time()     Returns the estimated time for a real run of the external 
#                                    polarisation offset program. This is obtained by calling the program
#                                    with the 'test' argument. Notice that if the program is not designed to 
#                                    honor this argument, undefined behaviour is to be expected!
# $id->get_impedance_frequencies($)  Function calculating the impedance frequencies and integration time 
#                                    for an impedance run the configuration file for the specified 
#                                    session number If no session number is specified, the current 
#                                    session is used by default the return is an aray of hashes containing 
#                                    frequency, integration time and resistor value If a second and third 
#                                    argument is specified, it is assumed to be a potentiostat object 
#                                    and a FRA object. In this case the frequency list will also include 
#                                    those tags which is defined in the frequency_seg_tags() lists. 
#                                    If the tag is not defined in the segment, the default value is used 
#                                    (from the impedance_potantiostat or impedance_FRA segments respectively
# $id->get_measure_nr()              Returns the current measure number.
# $id->increase_measure_nr()         Increases the current measure number.
# $id->increase_session_nr($)        Increases the session number (creates a new session). The sesion name of 
#                                    the new session is the specified string.
# $id->get_program_nr()              Returns the current program number.
# $id->increase_program_number()     Increases the current program number.
# $id->get_session_name($)           Returns the session name for the specified session, default is current session.
# $id->get_session_log($)            Returns the content of the sesion log file for the specified sesion, 
#                                    default is current session.
# $id->get_config_value($$)          Returns the value of the specified sesion and key (refer ElchemeaConfig.pm for details).
#                                    Arguments: section, key.
#                                    Returns an array of values if called in a list context (raw value split along commas).
# $id->config_section_exists($)      Checks if the specified section exists.
# $id->config_key_exists($$)         Checks if the specified key exists in the specified section. Arguments: section, key
# $id->change_config_value($$$)      Sets the value fo the specified section and key to the specified argument. Arguments:
#                                    section, key, newval.
# $id->name()                        Returns the user name.
# $id->session()                     Returns the current session number.
# $id->proglog($)                    Appends the specified string to the proglog file.                        
# $id->load_session_nr()             Loads the current session number from the test.nr file.
# $id->config()                      Returns an ElchemeaConfig instance tied to the current configuration file (for the current session).
#
#
################################################################################


use strict;
use warnings;
use ElchemeaHeader;
use ElchemeaConfig;
use Elchemea qw();
use Sys::Hostname;
use File::Copy;
use Fcntl qw(:DEFAULT :flock);
use Debug;
use SemaforeFile;

our @ISA = qw(Debug);

my $lock = '/tmp/ElchemeaUserlog.lock';

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
    ### make sure that lockfile exists and is writable by all ### 
    $lock = '/tmp/ElchemeaUserlog.lock';
    unless (-e $lock)
    {
        sysopen(FH, $lock, O_WRONLY | O_CREAT) or CORE::die "Error: Could not open $lock $!";
        chmod 0666, $lock;
        close FH;
    }
}


sub new
{
    my $class = shift;
    my $self = $class->SUPER::new();
    $self->{'_name'} = undef;
    $self->{'_session'} = 0;
    $self->{'_config'} = [];
    $self->{'_programnr'} = undef;
    $self->{'_programfile'} = undef;
    $self->{'_measurenr'} = undef;
    $self->{'_measurefile'} = undef;
    $self->{'_userlog'} = undef;
    $self->{'_sessionname'} = '';
    my $tn = shift;
    if ($tn =~ /(\w+)/)
    {
	$self->{'_name'} = $1;
    }
    bless $self,$class;
    $self->die("user '".$self->name."' does not exist!") unless Elchemea::user_exist($self->name);
    $self->init;
    return $self;
}

return 1;

sub modtime($)
{
    my $self = shift;
    my $session = $self->session;
    $session = shift if (@_);
    return $self->{'_config'}[$session]->modtime;
}


sub debug()
{
    my $self = shift;
    $self->SUPER::debug(@_);
    if (@_)
    {
	#### Pass on debug level to all objects honoring this ###
	foreach (@{$self->{'_config'}})
	{
	    $_->debug($self->{'_debug'});
	}
	$self->{'_measurefile'}->debug($self->{'_debug'});
	$self->{'_programfile'}->debug($self->{'_debug'});
	$self->{'_userlog'}->debug($self->{'_debug'});
    }
    return $self->{'_debug'};
}


sub init()
{
    my $self = shift;
    $self->load_session_nr;
    $self->{'_config'} = [];
    $self->writedebug("Initialising user '$self->{'_name'}' configuration") if ($self->debug);
    for my $x (0..$self->session)
    {
	my $fn = $userdir.$self->name.'/'.$x.'/config';
	$self->{'_config'}[$x] = ElchemeaConfig->new($fn);
	$self->{'_config'}[$x]->debug($self->{'_debug'});
    }
    $self->writedebug("Initialising measure and program number files") if ($self->debug);
    $self->{'_measurefile'} = SemaforeFile->new($userdir.$self->name.'/'.$self->session.'/measure.nr',$lock);
    $self->{'_programfile'} = SemaforeFile->new($userdir.$self->name.'/'.$self->session.'/program.nr',$lock);
    $self->writedebug("Loading measure and program number") if ($self->debug);
    $self->get_measure_nr;
    $self->get_program_nr;
    $self->{'_userlog'} = SemaforeFile->new($userdir.$self->name.'/'.$self->session.'/proglog.txt',$lock);
    $self->writedebug("Loading session name") if ($self->debug);
    my $namefile = SemaforeFile->new($userdir.$self->name.'/'.$self->session.'/name.txt',$lock);
    my @lines = $namefile->readlines();
    if (@lines && defined($lines[0]))
    {
	$self->{'_sessionname'} = $lines[0];
    }
}

sub get_channels
{
    my $self = shift;
    my @temps;
    my @stemps;
    my @volts;
    my $spl   = Elchemea::get_config_value('global','splineinterpol').' '.Elchemea::get_config_value('global','splineinterpoldata');
    my $tstr = $self->get_config_value('datalog','channels_temp');
    my $vstr = $self->get_config_value('datalog','channels_voltage');
    my $ststr = $self->get_config_value('datalog','channels_simpletemp');
    @temps = split(',',$tstr) if ($tstr);
    @stemps = split(',',$ststr) if ($ststr);
    @volts = split(',',$vstr) if ($vstr);
    my $cjid = $self->get_config_value('datalog','cj_channel');
    my @vres = ();
    my @stres = ();
    my @tres = ();
    my @error = ();
    foreach my $name (@stemps)
    {
	my $chid = $self->get_config_value('datalog','channel_'.$name);
	next unless ($chid);
	$chid =~ s/\,.*//;
	my $r = Elchemea::socket_client('localhost','12345','K',$chid);
	if ($r =~ /([\d\.\-\+eE]+)[\s\t]*$/)
	{
	    push @stres, $name.': '.$1;
	}
	else
	{
	    push @error,'Simple temperature "'.$name.'": '.$r;
	}
    }
    foreach my $name (@volts)
    {
	my $chid = $self->get_config_value('datalog','channel_'.$name);
	next unless ($chid);
	$chid =~ s/\,.*//;
	my $r = Elchemea::socket_client('localhost','12345','K',$chid);
	if ($r =~ /([\d\.\-\+eE]+)[\s\t]*$/)
	{
	    push @vres, $name.': '.$1;
	}
	else
	{
	    push @error,'Simple voltage "'.$name.'": '.$r;
	}
    }
    my $cjr = undef;
    my $cjtype = undef;
    if ($cjid)
    {
	my $chid;
	($chid,$cjtype) = split(/\s*,\s*/,$cjid);
	next unless ($chid);
	my $r = Elchemea::socket_client('localhost','12345','K',$chid);
	if ($r =~ /([\d\.\-\+eE]+)[\s\t]*$/)
	{
	    $cjr = $1;
	}
	else
	{
	    push @error,'Cold junction "'.$cjid.'": '.$r;
	}
    }
    foreach my $name (@temps)
    {
	my $val;
	my $setup = $self->get_config_value('datalog','channel_'.$name);
	my ($chid,$type) = split(/\s*,\s*/,$setup);
	next unless ($chid);
	my $r = Elchemea::socket_client('localhost','12345','K',$chid);
	if ($r =~ /([\d\.\-\+eE]+)[\s\t]*$/)
	{
	    $val = $1;
	    if (!$type)
	    {
		push @vres, $name.': '.$val;
		next;
	    }
	}
	else
	{
	    push @error,'Temperature "'.$name.'": '.$r;
	}
	next unless ($cjr && $spl && $spl =~ /\w+/);
	my $lead_res = $self->get_config_value('datalog','lead_resistance');
	$lead_res = 0 unless ($lead_res);
        my $deltaohm = 0;
        if ($cjtype eq 'pt1000')
        {
            ## convert to pt100 ###
            $deltaohm = ($cjr - $lead_res) / 10.0;
        }
        if ($cjtype eq 'pt100')
        {
            $deltaohm = ($cjr - $lead_res);
        }
        ### Convert pt100 to mV for thermocouple ####
        my $cmd = $spl.'/pt100-res-to-type-'.$type.'-mV '.$deltaohm;
        my $cjv = `$cmd`;
        $cjv =~ /([\d\.\-\+eE]+)/;
        $cjv = $1;
        ## Add cold junction voltage to thermovoltage ###
        my $samplev = $cjv + $val;
        ### Convert thermovoltage to temperature ####
        $cmd = $spl.'/NIST-90-'.$type.'-mV-to-temp '.$samplev;
        my $temp = `$cmd`;
        if ($temp =~ /([\d\.\-\+eE]+)/)
        {
            $temp = $1;
        }
        else
        {
            $temp = -32768;
        }
	push @tres, $name.': '.$temp;
    }
    return join(', ',@vres,@tres,@stres,@error);
}


sub log_string
{
    my $self = shift;
    my $res = '';
    my $session = $self->session;
    my $sticky = 0;
    $self->writedebug("Getting last userlog entry") if ($self->debug > 1);
    my $cv = $self->get_config_value('measure','sticky_logfile_entires');
    $sticky = 1 if ($cv && $cv =~ /yes/i);
    $self->{'_userlog'}->open_readback();
    my $count = 0;
    my $logline;
    while ((my $logline = $self->{'_userlog'}->readline) && ($sticky || $count < 2))
    {
	$count++;
	next unless ($logline =~ /\#\s*(.*)$/);
	$res = $1;
	last;
    }
    $self->{'_userlog'}->close;
    return $res;
}

sub sample_name()
{
    my $self = shift;
    my $res = '';
    my $session = $self->session;
    $self->writedebug("getting sample name from userlog") if ($self->debug > 1);
    $self->{'_userlog'}->open_read;
    while (my $s = $self->{'_userlog'}->readline && !$res)
    {
	next unless ($s =~ /sample|Description/i );
	if ($s =~ /\=\s*(.*)$/)
	{
	    $res = $1;
	}
	last;
    }
    $self->{'_userlog'}->close;
    return $res;
}


sub get_measure_file($)
{
    my $self = shift;
    my $postfix = shift;
    $self->writedebug("Getting current measure filename") if ($self->debug > 1);
    return $userdir.$self->name.'/'.$self->session.'/s'.$self->session.'_'.$self->get_measure_nr.$postfix; 
}

sub fisher_yates_shuffle {
    my $self = shift;
    srand();
    my $array = shift;
    my $i;
    for ( $i = @$array ; --$i ; ) {
        my $j = int rand( $i + 1 );
        @$array[ $i, $j ] = @$array[ $j, $i ];
    }
}

sub get_potsweep_points($)
{
    my $self = shift;
    my $session = shift;
    $session = $self->session unless ($session);
    $self->writedebug("Loading potential sweep points from session '$session' for user '$self->{'_name'}'") if ($self->debug > 1);
    my @list = ();
    for (1..4)
    {
	my $bias = $self->get_config_value('potsweep','bias_point_'.$_,$session);
	if (defined($bias))
	{
	    $bias += 0;
	    push @list,$bias;
	}
    }
    return @list;
}

sub get_offset_program_time($)
{
    my $self = shift;
    my $mode = shift;
    my $session = shift;
    $session = $self->session unless ($session);
    my $time = 0;
    if ($mode eq 'impedance')
    {
	my $prog = $self->get_config_value('impedance_FRA','polarisation_offset_program',$session);
	if ($prog)
	{
	    my $offset_prog = Elchemea::get_external_polarisation_function($prog);
	    $time +=  ($offset_prog->exec('test'))[0] if (ref($offset_prog) eq 'RemoteExec' && $offset_prog->exists);
	}
	$prog = $self->get_config_value('impedance_potentiostat','polarisation_offset_program',$session);
	if ($prog)
	{
	    my $offset_prog = Elchemea::get_external_polarisation_function($prog);
	    $time +=  ($offset_prog->exec('test'))[0] if (ref($offset_prog) eq 'RemoteExec' && $offset_prog->exists);
	}
    }
    elsif ($mode eq 'chrono')
    {
	my $prog = $self->get_config_value('chrono','polarisation_offset_program',$session);
	if ($prog)
	{
	    my $offset_prog = Elchemea::get_external_polarisation_function($prog);
	    $time +=  ($offset_prog->exec('test'))[0] if (ref($offset_prog) eq 'RemoteExec' && $offset_prog->exists);
	}
    }
    elsif ($mode eq 'potsweep')
    {
	my $prog = $self->get_config_value('potsweep','polarisation_offset_program',$session);
	if ($prog)
	{
	    my $offset_prog = Elchemea::get_external_polarisation_function($prog);
	    $time +=  ($offset_prog->exec('test'))[0] if (ref($offset_prog) eq 'RemoteExec' && $offset_prog->exists);
	}
    }
    return $time;
}


sub get_impedance_frequencies
{
#### Function calculating the impedance frequencies and integration time for an impedance run  ####
#### the configuration file for the specified session number                                   ####
#### If no session number is specified, the current session is used by default                 ####
#### the return is an aray of hashes containing frequency, integration time and resistor value ####
#### If a second and third argument is specified, it is assumed to be a potentiostat object    ####
#### and a FRA object. In this case the frequency list will also include those tags which is   ####
#### defined in the frequency_seg_tags() lists. If the tag is not defined in the segment, the  ####
#### default value is used (from the impedance_potantiostat or impedance_FRA segments          ####
#### respectively                                                                              ####
    my $self = shift;
    my $session = shift;
    $session = $self->session unless ($session);
    my $pot = shift;
    my $fra = shift;
    $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address')) unless ($pot);
    $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address')) unless ($fra);
    $self->writedebug("Loading impedance ferequencies from session '$session' for user '$self->{'_name'}'") if ($self->debug > 1);
    my $y = 0;
    my $sweep_seg = $self->get_config_value('impedance_FRA','sweep_segments',$session);
    my $npd = $self->get_config_value('impedance_FRA','number_of_points_pr_decade',$session);
    my $int_mode = $self->get_config_value('impedance_FRA','integration_type',$session);
    my @arr = (1..$sweep_seg);
    @arr = reverse(@arr) if ($self->get_config_value('impedance_FRA','frequency_order',$session) =~ /reverse/i);
    my %pot_vars = ();
    if ($pot)
    {
	my @pot_tags = $pot->frequency_segment_tags();
	foreach (@pot_tags)
	{
	    my $val = $self->get_config_value('impedance_potentiostat',$_);
	    $val = $pot->get_default($_) unless defined($val);
	    $pot_vars{$_} = $val;
	}
    }
    my %fra_vars = ();
    if ($fra)
    {
	my @fra_tags = $fra->frequency_segment_tags();
	foreach (@fra_tags)
	{
	    my $val = $self->get_config_value('impedance_FRA',$_);
	    $val = $fra->get_default($_) unless defined($val);
	    $fra_vars{$_} = $val;
	}
    }
    my @list = ();
    foreach my $x (@arr)
    {
        my $minfreq = $self->get_config_value('impedance_FRA','minimum_frequency_'.$x,$session);
        my $maxfreq = $self->get_config_value('impedance_FRA','maximum_frequency_'.$x,$session);
        my $int_time = $self->get_config_value('impedance_FRA','integration_time_'.$x,$session);
        my $points = $self->get_config_value('impedance_FRA','number_of_points_'.$x,$session);
	my %seg = ();
	foreach my $key (keys %pot_vars)
	{
	    $seg{$key} = $self->get_config_value('impedance_FRA',$key.'_'.$x,$session);
	    $seg{$key} = $pot_vars{$key} unless ($seg{$key});
	}
	foreach my $key (keys %fra_vars)
	{
	    $seg{$key} = $self->get_config_value('impedance_FRA',$key.'_'.$x,$session);
	    $seg{$key} = $fra_vars{$key} unless ($seg{$key});
	}
        $points = $npd unless ($points);
        my $lminf = log($minfreq) / log(10);
        my $lmaxf = log($maxfreq) / log(10);
        my $n = ($lmaxf - $lminf) * $points;
        my $np = 1/$points;
        for (my $y = $lmaxf; $y > $lminf ; $y= $y - $np)
        {
            my $freq = 10**$y;
            $freq = sprintf("%.6f",$freq);
            $freq = sprintf("%.4f",$freq) if ($freq > 0.01);
            $freq = sprintf("%.4f",$freq) if ($freq > 0.1);
            $freq = sprintf("%.3f",$freq) if ($freq > 1);
            $freq = sprintf("%.2f",$freq) if ($freq > 10);
            $freq = sprintf("%.1f",$freq) if ($freq > 100);
            $freq = sprintf("%.0f",$freq) if ($freq > 1000);
            my $ltime = $int_time / $freq;
            my $mintime = 1 / $freq;
            if ($int_mode =~ /sec/)
            {
                $ltime = $int_time;
                if ($ltime < $mintime)
                {
                    $ltime = $mintime;
                }
            }
            if ($ltime < 1){$ltime = 1};
            if ($ltime =~ /\d+\.\d+/)
            {
                $ltime =~ /(\d+)/;
                $ltime = $1+1;
            }
            my %h = (
		     'frequency' => $freq,
		     'time' => $ltime,
		     'integration_number' => $int_time,
		     'integration_mode' => $int_mode,
		     'segment' => $x,
		     );
	    foreach my $k (keys %seg)
	    {
		$h{$k} = $seg{$k};
	    }
            push @list, \%h;
        }
    }
    if ($self->get_config_value('impedance_FRA','frequency_order',$session) =~ /reverse/i)
    {
        return reverse(@list);
    }
    if ($self->get_config_value('impedance_FRA','frequency_order',$session) =~ /random/i)
    {
        $self->fisher_yates_shuffle(\@list);
    }
    return @list;
}



sub get_measure_nr()
{
    ### Do not use cashed values for this as other processes may increase the measure number! ###
    my $self = shift;
    if (@_)
    {
	### If additional argument, session number is specified, and then look in old session ###
	my $session = shift;
	my $measurefile = SemaforeFile->new($userdir.$self->name.'/'.$session.'/measure.nr',$lock);
	my @lines = $measurefile->readlines();
	if (@lines && defined($lines[0]) && $lines[0] =~ /(\d+)/)
	{
	    return sprintf("%04d",$1);
	}
	return '0000';
    }
    $self->writedebug("Loading current measure number") if ($self->debug);
    $self->{'_measurenr'} = '0000';
    if ($self->{'_measurefile'}->exist)
    {
	my @lines = $self->{'_measurefile'}->readlines();
	if (@lines && defined($lines[0]) && $lines[0] =~ /(\d+)/)
	{
	    $self->{'_measurenr'} = sprintf("%04d",$1);
	}
    }
    return $self->{'_measurenr'};
}

sub increase_measure_nr()
{
    my $self = shift;
    $self->writedebug("Incrementing current measure number") if ($self->debug);
    unless ($self->{'_measurefile'}->exist)
    {
	$self->{'_measurefile'}->writeline('0000'); 
	$self->{'_measurenr'} = '0000';
    }
    else
    {
	my $nr = $self->get_measure_nr;
	$nr ++;
	$self->{'_measurenr'} = sprintf("%04d",int($nr));
	$self->{'_measurefile'}->writeline($self->{'_measurenr'});
    }
    return $self->{'_measurenr'};
}


sub increase_session_nr($)
{
    my $self = shift;
    my $tsession_name = shift;
    $self->writedebug("Incrementing session number") if ($self->debug);
    $tsession_name =~ /([\w\s\d\.\-\+]+)/;
    my $session_name = $1;
    sysopen(FH, $lock, O_WRONLY | O_CREAT) or $self->die("Could not open $lock $!");
    flock(FH, LOCK_EX);
    my $user = $self->name;
    my $file = $userdir.$user.'/test.nr';
    my $test = $self->session;
    my $otest = $test;
    $test++;
    open MYOUT,'>',$file or $self->die("Could not open file $file $!");
    print MYOUT $test,"\n";
    close MYOUT;
    $self->writedebug("Setting measure number file") if ($self->debug);
    my $testdir = $userdir.$user.'/'.$test;
    mkdir $testdir, 0777 unless -d $testdir;
    $file = $userdir.$user.'/'.$test.'/measure.nr';
    my $filenumber = 0;
    $filenumber = shift if (@_);
    open MYOUT,'>',$file or $self->die("Could not open file $file $!");
    print MYOUT sprintf("%04d\n",int($filenumber));
    close MYOUT;
    chmod 0666,$file;
    $self->writedebug("Setting program number file") if ($self->debug);
    $file = $userdir.$user.'/'.$test.'/program.nr';
    open MYOUT,'>',$file or $self->die("Could not open file $file $!");
    print MYOUT "0\n";
    close MYOUT;
    chmod 0666,$file;
    my $source = $userdir.$user.'/'.$otest.'/config';
    ### if the session number is less than 1 (in case of a new user etc.) ###
    ### The source configuration file is the master file                  ###
    $source = $userdir.$user.'/config' if ($otest < 1);
    my $target = $userdir.$user.'/'.$test.'/config';
    $self->writedebug("Copying configuration file to new position '$target'") if ($self->debug);
    copy($source,$target);
    chmod 0666,$target;
    $target =  $userdir.$user.'/'.$test.'/name.txt';
    open MYOUT,'>',$target or $self->die("Could not open file $target $!");
    print MYOUT $session_name,"\n";
    close MYOUT;
    flock(FH,LOCK_UN);
    close FH;
    $self->init();
    my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
    my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));
    my $host = hostname;
    my $string = 'Session '.$self->session.' started on '."$host\n";
    $string .= "\nSession name: $session_name\n";
    $string .= "Start file number: $filenumber\n" if ($filenumber); 
    $string .= "Potentiostat: ".$pot->type."\n";
    $string .= "Frequency response analyser: ".$fra->type."\n";
    $string .= "\n";
    $self->proglog($string);
    return $self->session;
}

sub get_program_nr()
{
    my $self = shift;
    return $self->{'_programnr'} if (defined($self->{'_programnr'}));
    $self->writedebug("Loading program number") if ($self->debug > 1);
    $self->{'_programnr'} = '00'; 
    if ($self->{'_programfile'}->exist)
    {
	my @lines = $self->{'_programfile'}->readlines();
	if (@lines && defined($lines[0]) && $lines[0] =~ /(\d+)/)
	{
	    $self->{'_programnr'} = sprintf("%02d",$1);
	}
    }
    return $self->{'_programnr'};
}

sub increase_program_nr()
{
    my $self = shift;
    $self->writedebug("Incrementing program number") if ($self->debug > 1);
    unless ($self->{'_programfile'}->exist)
    {
	$self->{'_programfile'}->writeline('0000'); 
	$self->{'_programnr'} = '0000';
    }
    else
    {
	my $nr = $self->get_program_nr;
	$nr++;
	$self->{'_programnr'} = sprintf("%04d",int($nr));
	$self->{'_programfile'}->writeline($self->{'_measurenr'});
    }
    return $self->{'_programnr'};
}

sub get_session_name()
{
    my $self = shift;
    return $self->{'_sessionname'} unless (@_);
    my $session = shift;
    if ($session =~ /(\d+)/)
    {
	my $s = $1;
	my $file = SemaforeFile->new($userdir.$self->name.'/'.$s.'/name.txt');
	my @list = $file->readlines();
	return $list[0] if (@list);
    }
}

sub get_sessionlog()
{
    my $self = shift;
    my $session = shift;
    $session = $self->session unless ($session);
    if ($session == $self->session)
    {
	return join("\n",$self->{'_userlog'}->readlines) if ($self->{'_userlog'}->exist);
    }
    else
    {
	my $file = SemaforeFile->new($userdir.$self->name.'/'.$session.'/proglog.txt',$lock);
	return join("\n",$file->readlines) if ($file->exist);
    }
    return '';
}

sub config_key_exists($$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    return $self->{'_config'}[$self->session]->exists($section,$key);
}

sub config_section_exists($)
{
    my $self = shift;
    my $section = shift;
    return $self->{'_config'}[$self->session]->section_exists($section);
}

sub get_config_value($$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    my $session = shift;
    $session = $self->session unless ($session);
    if (wantarray)
    {
	return ($self->{'_config'}[$session]->get_config_value($section,$key));
    }
    return $self->{'_config'}[$session]->get_config_value($section,$key);
}

sub change_config_value($$$)
{
    my $self = shift;
    my $section = shift;
    my $key = shift;
    my $newvalue = shift;
    return $self->{'_config'}[$self->session]->change_config_value($section,$key,$newvalue);
}

sub name()
{
    my $self = shift;
    return $self->{'_name'};
}

sub session()
{
    my $self = shift;
    return $self->{'_session'};
}

sub proglog($)
{
    my $self = shift;
    my $string = shift;
    chomp $string;
    $self->{'_userlog'}->lock_ex;
    unless ($self->{'_userlog'}->exist)
    {
	$self->writedebug("Writing to new userlog for session ".$self->session." for user '".$self->name."'") if ($self->debug);
	$self->{'_userlog'}->writeline();
	chmod 0666,$self->{'_userlog'}->filename;
    }
    ### Opens the logfile #####
    my $time = localtime;
    my $s = $time.' : '.$string;
    $self->{'_userlog'}->append($s);
    $self->{'_userlog'}->unlock;
}


sub load_session_nr()
### Function returning the session number ###
{
    my $self = shift;
    my $file = $userdir.$self->{'_name'}.'/test.nr';
    $self->{'_session'} = 0;
    if (-e $file)
    {
	$self->writedebug("Loading session number from file '$file'") if ($self->debug);
        sysopen(FH, $lock, O_WRONLY | O_CREAT) or $self->die("Could not open lockfile $lock $!");
        flock(FH, LOCK_SH);
        open MYIN,'<',$file or $self->die("Could not open file $file $!");
        my $test = <MYIN>;
        close MYIN;
        flock(FH, LOCK_UN);
        close FH;
        if ($test =~ /(\d+)/)
        {
            $self->{'_session'} = $1;
        }
    }
    return $self->{'_session'};
}


sub config()
{
    my $self = shift;
    return $self->{'_config'}[$self->session];
}
