package Hioki;
#$Id: Hioki.pm,v 1.33 2013/08/27 12:45:16 sofc Exp $

############################################################
#
#    Package Hioki
#
# This module implements the FRA class for Hioki
# --{\textregistered} 
# 3533 and 3532 Component Measuring Instruments.
#
# To get a device instance call one of the constructors:
#
# $fra = Hioki3522->new($gpib_address);
#
# $fra = Hioki3532->new($gpib_address);
#
############################################################


use strict;
use warnings;
use FDEV;
our @ISA = qw(FRA);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Hioki';
    $self->{'_virtual'} = 0;
    $self->{'_mode'} = 'V';
    $self->{'_modes'} = {'Volt' => 'V','Amp' => 'A', 'OCV' => 'O'},
    $self->{'_minf'} = 42;
    $self->{'_maxf'} = 100e3;
    $self->{'_minpV'} = 0;
    $self->{'_maxpV'} = 0;
    $self->{'_minpA'} = 0;
    $self->{'_maxpA'} = 0;
    $self->{'_maxampA'} = 0.1;
    $self->{'_maxampV'} = 5;
    $self->{'_overhead'} = 1.2;
    $self->{'_minintegration'} = 1;
    $self->{'_setuptime'} = 10;
    $self->{'_delay'} = 0; 
    $self->{'_freq_seg_tags'} = ['range'];
    $self->{'_range'} = undef;
    $self->{'_setup_tags'} = 'mode,overhead,amplitude,delay,monitor_enable,monitor_current_limit,monitor_voltage_limit,range,source_mode';
    $self->{'_desc'} = {
	'mode' => 'Control mode. If OCV is selected, the voltage conrol is made using the drive leads (not sense leads as in normal voltage control)',
	'overhead' => '',
	'amplitude' => 'AC amplitude (RMS)',
	'delay' => 'Delay in seconds before measurements at each frequency',
	'monitor_enable' => 'Enable current or voltage limit',
	'monitor_current_limit' => 'Maximum current if current limit is used',
	'monitor_voltage_limit' => 'Maximum voltage if voltge limit is used',
	'range' => 'Resistance range, can be either fixed or automatic. Note that in case of autoranging the instrument may go hunting',
	'source_mode' => 'Input mode',
    };
    $self->{'_values_source_mode'} = ['V/I'];
    $self->{'_values_range'} = ['auto',0.1,1,10,100,1000,1E4,1E5,1E6,1E7,1E8];
    $self->{'_errorcodes'} = {
        'S7' => 'Power on',
        'S6' => 'Unused',
        'S5' => 'Command error',
        'S4' => 'Execution error',
        'S3' => 'Device error/Short circuit',
        'S2' => 'Query error',
        'S1' => 'Unused',
        'S0' => 'Unused',
        'E07' => 'Unused',
        'E06' => 'Constant current/voltage overflow',
        'E05' => 'Limit overflow',
        'E04' => 'Impedance range overflow',
        'E03' => 'Impedance Range underflow',
        'E02' => 'Data sampling completed',
        'E01' => 'Measurement completed',
        'E00' => 'Compensation data measurement completed',
        'E17' => 'Unused',
        'E16' => 'Both parameters within limits',
        'E15' => 'Second parameter below limit',
        'E14' => 'Second parameter within limits',
        'E13' => 'Second parameter above upper limit',
        'E12' => 'First parameter below limit',
        'E11' => 'First parameter within limits',
        'E10' => 'First parameter upper upper limit',
        '' => ' ',
        'XX' => 'No data returned',
    };
    $self->{'_default'} = {
        'mode' => 'Volt',
        'source_mode' => 'V/I',
	'range' => 'auto',
        'monitor_enable' => 'NO',
	'monitor_current_limit' => 0.1,
	'monitor_voltage_limit' => 5,
        'external_polarisation' => 'OFF',
        'reference_potential' => 'Fixed',
        'delay' => $self->{'_delay'},
        'amplitude' => $self->{'_minampV'},
        'polarisation' => 0,
    };
    $self->{'_values_monitor_enable'} = ['YES','NO'];
    $self->{'_save_tags'} = [':MEAS:ITEM',':CORR:OPEN',':CORR:SHOR',':TRIG',':TRIG:DELA',':COMP',':SCAL',':HEAD',':PAR1',':PAR2',':PAR3',':PAR4',':RANGE:AUTO',':RANGE',':SPEED',':LEV',':LIM',':LIM:CURR',':LIM:VOLT',':LEV:CCURR',':LEV:CVOLT',':LEV:VOLT',':FREQ',':AVER'];
    bless $self,$class;
    return $self;
}


sub save()
{
    my $self = shift;
    my @res = ();
    foreach my $tag (@{$self->{'_save_tags'}})
    {
        push @res,$tag.' '.$self->readvalue($tag);
    }
    return @res;
}


sub readvalue($)
{
    my $self = shift;
    my $tag = shift;
    my $res = undef;
    my $count = 0;
    while (!defined($res) && $count++ < 5)
    {
	$self->writecmd($tag.'?');
	my $r = $self->readcmd;
	if ($r =~ /([\w\.\+\-\,eE]+)/)
	{
	    $res = $1;
	}
    }
    return $res if (defined($res));
    $count--;
    my $str = "Could not read value for tag '$tag', tried $count times!";
    $self->errorlog($str);
    return undef;
}

sub set_polarisation($)
{
    my $self = shift;
    my $pol = shift;
    ### Set output mode ###
    if ($self->{'_mode'} =~ /A/)
    {
	$self->writecmd(':LEV CC');
    }
    elsif ($self->{'_mode'} =~ /V/)
    {
	$self->writecmd(':LEV CV');
    }
    else
    {
	$self->writecmd(':LEV V');
    }
    ### Hioki can not set a bias polarisation, so always return 0 ###
    return 0;
}

sub setup_impedance()
{
    my $self = shift;
    my $setup = shift;
    $self->reset();
    $self->checkcmd(':MEAS:ITEM 5,18');
    $self->checkcmd(':CORR:OPEN OFF');
    $self->checkcmd(':BEEP:COMP OFF');
    $self->checkcmd(':BEEP:KEY OFF');
    $self->checkcmd(':CORR:SHOR OFF');
    $self->checkcmd(':TRIG EXT');
    $self->checkcmd(':TRIG:DELA 1');
    $self->checkcmd(':COMP OFF');
    $self->checkcmd(':SCAL OFF');
    $self->checkcmd(':HEAD OFF');
    $self->checkcmd(':PAR1 Z');
    $self->checkcmd(':PAR2 PHAS');
    $self->checkcmd(':PAR3 RS');
    $self->checkcmd(':PAR4 X');
    $self->checkcmd(':RANGE:AUTO ON');
    $self->checkcmd(':SPEED SLOW2');
    my $monitor = $self->get_setup_value($self->get_default('monitor_enable'),$setup,'monitor_enable');
    if ($monitor eq 'YES')
    {
	$self->writecmd(':LIM ON');
	if ($self->{'_mode'} eq 'A')
	{
	    my $mv = $self->get_setup_value(0.1,$setup,'monitor_current_limit');
	    $self->checkcmd(':LIM:CURR '.$mv);
	}
	else
	{
	    my $mv = $self->get_setup_value(5,$setup,'monitor_voltage_limit');
	    $self->checkcmd(':LIM:VOLT '.$mv);
	}
    }
    else
    {
	$self->checkcmd(':LIM OFF');
    }
    $self->set_polarisation(0); ### Set output mode ###
    $self->amplitude($self->get_setup_value(0,$setup,'amplitude'));
    $self->tag_value('range',$self->get_setup_value($self->get_default('range'),$setup,'range'));
    $self->{'_delay'} = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    $self->reset_data;
    return 0;
}

sub tag_value
{
    my $self = shift;
    my $key = shift;
    if ($key eq 'range')
    {
	return $self->{'_range'} unless (@_);
	my $val = shift;
	my @list = @{$self->{'_values_range'}};
        for (my $x = 0; $x < @list; $x++)
        {
	    next unless ($val eq $list[$x]);
	    if ($x)
	    {
		$self->checkcmd(':RANGE:AUTO OFF');
		$self->checkcmd(':RANGE '.$x);
	    }
	    else
	    {
		$self->checkcmd(':RANGE:AUTO ON');
	    }
	    $self->{'_range'} = $list[$x];
	    last;
	}
	return $self->{'_range'};
    }
}


sub amplitude
{
    my $self = shift;
    return $self->{'_amplitude'} unless (@_);
    my $amp = shift;
    if ($self->{'_mode'} eq 'A')
    {
        $amp = $self->{'_minampA'} if ($amp < $self->{'_minampA'});
        $amp = $self->{'_maxampA'} if ($amp > $self->{'_maxampA'});
        $self->checkcmd(sprintf(":LEV:CCURR %.4g",$amp));  ## Current AC amplitude ##
    }
    elsif ($self->{'_mode'} eq 'V')
    {
        $amp = $self->{'_minampV'} if ($amp < $self->{'_minampV'});
        $amp = $self->{'_maxampV'} if ($amp > $self->{'_maxampV'});
        $self->checkcmd(sprintf(":LEV:CVOLT %.4g",$amp));  ## Voltage AC amplitude ##
    }
    else
    {
	$amp = $self->{'_minampV'} if ($amp < $self->{'_minampV'});
	$amp = $self->{'_maxampV'} if ($amp > $self->{'_maxampV'});
        $self->checkcmd(sprintf(":LEV:VOLT %.4g",$amp));  #Voltage AC amplitude across
        # sample AND a 50 Ohm series impedance in output!
    }
    $self->{'_amplitude'} = $amp;
    return $self->{'_amplitude'};
}

sub errorcode($)
{
    my $self = shift;
    return $self->{'_errorcodes'}{$_[0]} if (exists($self->{'_errorcodes'}{$_[0]}));
    return $self->S0code(@_);
}

sub warningcode($)
{
    my $self = shift;
    return $self->S1code(@_);
}

sub Ecode($)
{
    my $self = shift;
    my $val = shift;
    return '' unless ($val);
    $val += 0;
    my @l = ();
    for my $x (0..7)
    {
        $l[$x] = 0;
        $l[$x] = 1 if ($val & 2**$x);
    }
    my @r = ();
    for my $x (0..7)
    {
        push @r, $self->{'_errorcodes'}{'S'.$x} if ($l[$x]);
    }
    my $res = join(', ',@r);
    return $res;
}

sub S0code($)
{
    my $self = shift;
    my $val = shift;
    return '' unless ($val);
    $val += 0;
    my @l = ();
    for my $x (0..7)
    {
        $l[$x] = 0;
        $l[$x] = 1 if ($val & 2**$x);
    }
    my @r = ();
    for my $x (0..7)
    {
        push @r, $self->{'_errorcodes'}{'E0'.$x} if ($l[$x]);
    }
    my $res = join(', ',@r);
    return $res;
}

sub S1code($)
{
    my $self = shift;
    my $val = shift;
    return '' unless ($val);
    $val += 0;
    my @l = ();
    for my $x (0..7)
    {
        $l[$x] = 0;
        $l[$x] = 1 if ($val & 2**$x);
    }
    my @r = ();
    for my $x (0..7)
    {
        push @r, $self->{'_errorcodes'}{'E1'.$x} if ($l[$x]);
    }
    my $res = join(', ',@r);
    return $res;
}

sub get_S0()
{
    my $self = shift;
    $self->writecmd('ESR0?');
    my $rep = $self->readcmd();
    if ($rep && $rep =~ /(\d+)/)
    {
        my $e = $1;
        my $r = $self->S0code($e);
        $self->writedebug("Event status register 0: $r") if ($r);
        if (@_)
        {
            my @l = ();
            for my $x (0..7)
            {
                $l[$x] = 0;
                $l[$x] = 1 if ($e & 2**$x);
            }
            $e = 0;
            for my $x (0..7)
            {
                $e += 2**$x if ($l[$x]);
            }
            return $e;
        }
        return $rep;
    }
    return '';

}

sub get_S1()
{
    my $self = shift;
    $self->writecmd('ESR1?');
    my $rep = $self->readcmd();
    if ($rep && $rep =~ /(\d+)/)
    {
        my $e = $1;
        my $r = $self->S0code($e);
        $self->writedebug("Event status register 1: $r") if ($r);
        if (@_)
        {
            my @l = ();
            for my $x (0..7)
            {
                $l[$x] = 0;
                $l[$x] = 1 if ($e & 2**$x);
            }
            $e = 0;
            for my $x (0..7)
            {
                $e += 2**$x if ($l[$x]);
            }
            return $e;
        }
        return $rep;
    }
    return '';

}



sub get_error()
{
    my $self = shift;
    $self->writecmd('*ESR?');
    my $rep = $self->readcmd();
    if ($rep && $rep =~ /(\d+)/)
    {
        my $e = $1;
        my $r = $self->Ecode($e);
        $self->writedebug("Standard event status error: $r") if ($r);
        if (@_)
        {
            my @l = ();
            for my $x (0..7)
            {
                $l[$x] = 0;
                $l[$x] = 1 if ($e & 2**$x);
            }
            # Specifically set $l[7] to 0 to ignore 'power on flag'#
            $l[7] = 0;
            $e = 0;
            for my $x (0..7)
            {
                $e += 2**$x if ($l[$x]);
            }
            return $e;
        }
        return $rep;
    }
    return '';
}


sub run_frequency()
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $setup = shift;
    my $testfile = undef;
    $testfile = $$setup{'semafore_file'} if (exists $$setup{'semafore_file'} && $$setup{'semafore_file'});
    return undef unless (!$testfile || -e $testfile);
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    my $avg = $self->get_avg($freq,$time);
    $self->checkcmd(sprintf(":FREQ %.4g",$freq));
    $self->checkcmd(':AVER '.$avg);
    my $delay = $self->get_setup_value($self->{'_delay'},$setup,'delay');
    sleep($delay);
    $self->writecmd('*TRG');
    $time = $self->get_integration_time($freq,$time);
    $self->writedebug("Estimated integration time: $time seconds");
    my $etime = time + $time;
    my $maxtime = $etime + $time;
    my $count = 1;
    while (time < $etime)
    {
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->();
        }
        next if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	return undef;
    }
  OUTER:
    while (time < $maxtime)
    {
        my $readloop = 0;
      INNER:
        while ($readloop < 5)
        {
	    $self->writecmd(':ESR0?');
	    my $r = $self->readcmd();
	    if ($r)
	    {
		$self->writedebug('Event status register 0: '.$self->errorcode($r));
		my $e = $r;
		my @l = ();
		for my $x (0..7)
		{
		    $l[$x] = 0;
		    $l[$x] = 1 if ($e & 2**$x);
		}
		### Explicitly set 1 and 2 to zero as they are not errors ###
		$l[1] = 0;
		$l[2] = 0;
		$e = 0;
		for my $x (0..7)
		{
		    $e += 2**$x if ($l[$x]);
		}
		$self->{'_last_error'} = $e;
	    }
	    if ($r =~ /^(\d+)/)
	    {
		$r = int($1+0);
		my $test = $r & 2;
                # Bitwise and of result and 2 to #
		# get bit 1 or result only #
		unless ($test)
		{
		    $readloop++;
		    next INNER;
		}
	    }
	    else
	    {
		$readloop++;
		next INNER;
	    }
	    $self->writecmd('MEAS?');
            my $res = $self->readcmd();
            if ($res =~ /MEAS/i)
            {
                my $result = sprintf("%f %f %f",$freq,0,0);
		push @{$self->{'_data'}},$result;
		return $result;
            }
            my @rep = split(",",$res);
	    my $z = $rep[0];
	    my $phas = $rep[1] * 2 * 3.141592 / 360;
	    my $real = $z * cos($phas);
	    my $imag = $z * sin($phas);
            my $result = sprintf("%f %f %f",$freq,$real,$imag);
	    push @{$self->{'_data'}},$result;
	    my $warn = $self->get_S1('ignore');
	    $self->{'_last_warning'} = $warn if ($warn);
	    return $result;
        }
        sleep(1);
	if (exists($$setup{'function'}) && $count++ > $self->{'_function_reload'})
        {
            $count = 1;
            $$setup{'function'}->() if (ref($$setup{'function'}) eq 'CODE');
        }
        next OUTER if (!$testfile || -e $testfile);
	$self->reset();
	push @{$self->{'_data'}},undef;
	return undef;
    }
    push @{$self->{'_data'}},undef;
    return undef;
}

sub run_frequency_all
{
    my $self = shift;
    my $result = $self->run_frequency(@_);
    $result .= ' #';
    my $r = $self->readvalue(':RANGE');
    $result .= ' Range: '.$self->{'_values_range'}[$r];
    push @{$self->{'_longdata'}},$result;
    return $result;
}

sub get_min_integration_time($)
{
    my $self = shift;
    my $freq = shift;
    $freq = $self->{'_minf'} if ($freq < $self->{'_minf'});
    $freq = $self->{'_maxf'} if ($freq > $self->{'_maxf'});
    my $avg = $self->get_avg($freq,1);
    my $t = POSIX::ceil($self->{'_delay'} + $avg * $self->get_n($freq) / $freq);
    $t = $self->{'_minintegration'} if ($t < $self->{'_minintegration'});
    return $t;
}

sub get_integration_time($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $avg = $self->get_avg($freq,$time);
    my $mtime = POSIX::ceil($self->{'_delay'} + $avg * $self->get_n($freq) / $freq);
    $time = $mtime if ($time < $mtime);
    my $mintime = $self->get_min_integration_time($freq);
    $time = $mintime if ($time < $mintime);
    return $time;
}

sub get_n()
{
    my $self = shift;
    my $f = shift;
    return 3 if ($f < 10);
    return 32 if ($f < 187);
    return 160 if ($f < 937.5);
    return 800 if ($f < 4687);
    return 4000 if ($f < 23430);
    return 20000;
}

sub get_avg($$)
{
    my $self = shift;
    my $freq = shift;
    my $time = shift;
    my $tout = $time * $freq / $self->get_n($freq);
    my $avg = int((log($tout)/log(2))+1);
    $avg = 2**$avg;
    $avg = 64 if ($avg > 64);
    $avg = 2 if ($avg < 2);
    return $avg;
}

sub init()
{
    ### NB Do not use '*RST' as this sets freq to 1kHz and OCV amplitude ##
    ### to 1 Volt!!! ###
    my $self = shift;
    $self->reset();
    $self->checkcmd(':MEAS:ITEM 5,18');
    $self->checkcmd(':CORR:OPEN OFF');
    $self->checkcmd(':BEEP:COMP OFF');
    $self->checkcmd(':BEEP:KEY OFF');
    $self->checkcmd(':CORR:SHOR OFF');
    $self->checkcmd(':TRIG EXT');
    $self->checkcmd(':TRIG:DELA 1');
    $self->checkcmd(':COMP OFF');
    $self->checkcmd(':SCAL OFF');
    $self->checkcmd(':HEAD OFF');
    $self->checkcmd(':PAR1 Z');
    $self->checkcmd(':PAR2 PHAS');
    $self->checkcmd(':PAR3 RS');
    $self->checkcmd(':PAR4 X');
    $self->checkcmd(':RANGE:AUTO ON');
    $self->checkcmd(':SPEED SLOW2');
    $self->writecmd('*idn?');
    return $self->readcmd();
}

sub isallive()
{
    my $self = shift;
    $self->writecmd(':TRAN:TERM 1');
    $self->writecmd('*idn?');
    my $res = $self->readcmd();
    return 1 if (defined($res) && $res =~ /hioki/i);
    return 0;
}

sub reset()
{
    ### NB Do not use '*RST' as this sets freq to 1kHz and OCV amplitude ##
    ### to 1 Volt!!! ###
    ### Instead park at 1kHz with amplitude 10 mV  ###
    my $self = shift;
    $self->writecmd('*CLS');
    $self->writecmd(':TRAN:TERM 1');
    $self->checkcmd(':LEV V;:LEV:VOLT 0.01');
    $self->checkcmd(':FREQ 1000');
}

sub checkcmd($)
{
    ########################################################
    #### checkcmd($)                                     ###
    #### runs command twice if first go results in error ###
    ########################################################
    my $self = shift;
    my $cmd = shift;
    $self->writecmd($cmd);
    $self->writecmd('*ESR?');
    my $e = $self->readcmd();
    if ($e && ($e+0))
    {
        $self->writecmd($cmd);
        $self->checkerror();
    }
}

sub checkerror()
{
    my $self = shift;
    $self->writecmd('*ESR?');
    my $e = $self->readcmd();
    if ($e && ($e+0))
    {
        $self->writedebug("ERROR cummunicating with $self->{'_type'}: message: '".$self->Ecode($e)."'");
        $self->errorlog("ERROR cummunicating with $self->{'_type'}: message: '".$self->Ecode($e)."'");
        $self->writecmd('*CLS');
    }
}

package Hioki3522;

our @ISA = qw(Hioki);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Hioki 3522';
    $self->{'_minf'} = 1e-3;
    $self->{'_maxf'} = 100e3;
    bless $self,$class;
    return $self;
}


package Hioki3532;

our @ISA = qw(Hioki);

sub new
{
    my $class = shift;
    my $addr = shift;
    my $self = $class->SUPER::new($addr,@_);
    $self->{'_type'} = 'Hioki 3532';
    $self->{'_minf'} = 42;
    $self->{'_maxf'} = 5e6;
    bless $self,$class;
    return $self;
}

return 1;


